#include <directfb.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "cst.h"
#include "cst_dfb_common.h"
#include "cst_card.h"
#include "cst_survey.h"
#include "cst_storage.h"

typedef unsigned int page_t;

enum {
	PAGE_TYPE_MASK		= 0xffff0000,
	PAGE_TYPE_INIT		= 0x00000000,
	PAGE_TYPE_START		= 0x00010000,
	PAGE_TYPE_SURVEY	= 0x00020000,
	PAGE_TYPE_END		= 0x00030000,
};

#define PAGESTACK_SIZE (64)

static page_t PageStackBuffer[PAGESTACK_SIZE];
static page_t *PageStack;

char *
CSTGetIPaddrString(void)
{
	static char str[16] = " ";
	struct ifreq ifr;
	static int init = 1;
	int fd;

	if (init) {
		fd = socket(AF_INET, SOCK_DGRAM, 0);
		if (fd == -1)
			return NULL;

		ifr.ifr_addr.sa_family = AF_INET;
		strcpy(ifr.ifr_name, "eth0");

		ioctl(fd, SIOCGIFADDR, &ifr);
		if (fd == -1)
			return NULL;

		close(fd);

		if (ifr.ifr_addr.sa_data[2] == 0xff ||
		    ifr.ifr_addr.sa_data[3] == 0xff ||
		    ifr.ifr_addr.sa_data[4] == 0xff ||
		    ifr.ifr_addr.sa_data[5] == 0xff)
			return str;

		sprintf(str, "%d.%d.%d.%d",
			(unsigned char)ifr.ifr_addr.sa_data[2],
			(unsigned char)ifr.ifr_addr.sa_data[3],
			(unsigned char)ifr.ifr_addr.sa_data[4],
			(unsigned char)ifr.ifr_addr.sa_data[5]);

		init = 0;

		CSTINFO("IPADDR(%s)\n", str);
	}

	return str;
}

char *
CSTGetMACaddrString(void)
{
	static char str[16];
	struct ifreq ifr;
	static int init = 1;
	int fd;

	if (init) {
		fd = socket(AF_INET, SOCK_DGRAM, 0);
		if (fd == -1)
			return NULL;

		ifr.ifr_addr.sa_family = AF_INET;
		strcpy(ifr.ifr_name, "eth0");

		ioctl(fd, SIOCGIFHWADDR, &ifr);
		if (fd == -1)
			return NULL;

		close(fd);

		sprintf(str, "%02X%02X%02X:%02X%02X%02X",
			(unsigned char)ifr.ifr_hwaddr.sa_data[0],
			(unsigned char)ifr.ifr_hwaddr.sa_data[1],
			(unsigned char)ifr.ifr_hwaddr.sa_data[2],
			(unsigned char)ifr.ifr_hwaddr.sa_data[3],
			(unsigned char)ifr.ifr_hwaddr.sa_data[4],
			(unsigned char)ifr.ifr_hwaddr.sa_data[5]);

		init = 0;

		CSTINFO("MACADDR(%s)\n", str);
	}

	return str;
}

static char FileLabel[FILENAME_MAX];

void
CSTSetFileLabel(char *label)
{
	if (label)
		strncpy(FileLabel, label, FILENAME_MAX);
	else
		FileLabel[0] = '\0';
}

char *
CSTGetFileLabel(void)
{
	return FileLabel;
}

char *
CSTRequestLabel(void)
{
	static int init = 1;
	char *label;
	int ret;

	if (init) {
		srand(time(NULL));
		init = 0;
	}

	label = (char *)malloc(FILENAME_MAX);
	if (!label)
		return NULL;

	while(1) {
		char file[FILENAME_MAX];
		char *macaddr = CSTGetMACaddrString();
		char id[FILENAME_MAX];

		sprintf(id, "ID_%6s_%d",&macaddr[7], rand());
		CSTINFO("ID(%s)\n", id);

		sprintf(label, STORAGE_DIR"%s", id);

		sprintf(file, "%s.txt", label);
		ret = CSTStorage_CheckFile(file);
		if (ret == 1)
			continue;

		sprintf(file, "%s.jpg", label);
		ret = CSTStorage_CheckFile(file);
		if (ret == 1)
			continue;

		sprintf(label, WORK_DIR"%s", id);
		break;
	}

	return label;
}

void
CSTReleaseLabel(char *label)
{
	free(label);
}

static void
PageStackInit(void)
{
	int i;
	for (i=0; i< PAGESTACK_SIZE; i++)
		PageStackBuffer[i] = 0;

	PageStack = &PageStackBuffer[PAGESTACK_SIZE];
}

static void
PageStackPush(page_t id)
{
	*(--PageStack) = id;
}

static page_t
PageStackPop(void)
{
	return *(PageStack++);
}

static void
PageStackDump(void)
{
#if defined(CST_DEBUG_PAGESTACK)
	unsigned int *stack = PageStack;
	while (stack <= &PageStackBuffer[PAGESTACK_SIZE - 1])
		CSTDEBUG("PageStack(%08x)\n", *(stack++));
#endif
}

static void
CSTSurveyDataReset(void)
{
	int sheet_id = CSTSurvey_GetTopSheetID();
	SSheet_t *sheet = CSTSurvey_GetSheetInfo(sheet_id);
	int i;

	while (sheet) {
		for (i=0; i<sheet->nr_answer; i++)
			sheet->answer[i].result = 0;

		sheet = sheet->next;
	}
}

static int
CSTSurveySaveResult(void)
{
	int sheet_id = CSTSurvey_GetTopSheetID();
	SSheet_t *sheet = CSTSurvey_GetSheetInfo(sheet_id);
	FILE *fp = stdout;
	int page = 1;
	int i;

	time_t t;
	struct tm *tm_time;
	char timebuf[64];

	char file[FILENAME_MAX];
	char *id;

	CSTFUNC();

	t = time(NULL);
	tm_time = localtime(&t);
	strftime(timebuf, 64, "%Y/%m/%d_%H:%M:%S", tm_time);

	sprintf(file, "%s.txt", CSTGetFileLabel());
	fp = fopen(file, "w");

	fprintf(fp, "%s,", timebuf);

	id = CSTGetFileLabel();
	id = strstr(id, "ID_");
	fprintf(fp, "%s,", id);

	while (sheet) {
		fprintf(fp, "[%d],", page++);
		for (i=0; i<sheet->nr_answer; i++)
			fprintf(fp, "%d,", sheet->answer[i].result);

		sheet = sheet->next;
	}

	fprintf(fp, "\n");
	fclose(fp);

	return CSTStorage_SaveFile(file);  
}

static void
CSTCleanWorkDir(void)
{
	char cmd[FILENAME_MAX];

	sprintf(cmd, "rm -f " WORK_DIR "*.jpg " WORK_DIR "*.txt");
	system(cmd);
	CSTDEBUG("CMD(%s)\n", cmd);
}

int
main(int argc, char *argv[])
{
	int flag_exit = 0;
	page_t page_id;
	int sheet_id;
	CSTProcResult_t result;
	int ret;

	CSTINFO("Start!\n");
	ret = CSTSurvey_ParseFile(RESOURCE_DIR"survey.txt");
	if (ret) {
		CSTERR("File not found. ("RESOURCE_DIR"survey.txt)\n");
		return 1;
	}

	CSTDirectFBSetup(argc, argv);

	CSTStorage_Init();
	CSTCard_Start();

	PageStackInit();

	page_id = (PAGE_TYPE_INIT | PROC_PAGE_START); /* Top Page ID */

	CSTStartWindow_PrepareResource();
	CSTSurveyWindow_PrepareResource();
	CSTEndWindow_PrepareResource();

	while (!flag_exit) {
		PageStackPush(page_id);
		PageStackDump();

		switch (page_id & PAGE_TYPE_MASK) {
		case PAGE_TYPE_INIT:
			page_id = (PAGE_TYPE_START | PROC_PAGE_START);
			break;
		case PAGE_TYPE_START:
			sheet_id = (page_id & ~PAGE_TYPE_MASK);

			if (sheet_id == PROC_PAGE_START) {
				CSTINFO("---\n");
				CSTCleanWorkDir();
				CSTSurveyDataReset();
				CSTSetFileLabel(NULL);
			}

			CSTStartWindow_Create(sheet_id);
			CSTStartWindow_Proc(sheet_id, &result);
			CSTDEBUG("ProcResult(ID: %08x type: %04x,"
				 "page: %04x, data: %p)\n",
				 page_id, result.type,
				 result.page, result.data);

			switch (result.type) {
			case PROC_RESULT_NEXT:
				if (result.page == PROC_PAGE_FINISH) {
					char *label = CSTGetFileLabel();
					if (label[0] == '\0') {
						label = CSTRequestLabel();
						CSTSetFileLabel(label);
						CSTReleaseLabel(label);
					}
					page_id = (PAGE_TYPE_SURVEY |
						   CSTSurvey_GetTopSheetID());
				} else
					page_id = (PAGE_TYPE_START |
						   result.page);
				break;

			case PROC_RESULT_CANCEL:
				PageStackPop();
				page_id = PageStackPop();
				break;

			case PROC_RESULT_EXIT:
				flag_exit = 1;
				break;
			default:
				break;
			}
			break;
		case PAGE_TYPE_SURVEY:
			sheet_id = (page_id & ~PAGE_TYPE_MASK);
			CSTSurveyWindow_Create(sheet_id);
			CSTSurveyWindow_Proc(sheet_id, &result);
			CSTDEBUG("ProcResult(ID: %08x type: %04x, "
				 "page: %04x, data: %p)\n",
				 page_id, result.type,
				 result.page, result.data);

			switch (result.type) {
			case PROC_RESULT_NEXT:
				if (result.page == PROC_PAGE_FINISH)
					page_id = (PAGE_TYPE_END |
						   PROC_PAGE_START);
				else
					page_id = (PAGE_TYPE_SURVEY |
						   result.page);
				break;

			case PROC_RESULT_CANCEL:
				PageStackPop();
				page_id = PageStackPop();
				break;

			case PROC_RESULT_ALLCANCEL:
				page_id = (PAGE_TYPE_START | PROC_PAGE_START);

				PageStackInit();
				break;

			default:
				break;
			}
			break;
		case PAGE_TYPE_END:
			CSTEndWindow_Create();

			ret = CSTSurveySaveResult();

			CSTEndWindow_Wait(ret);

			page_id = (PAGE_TYPE_START | PROC_PAGE_START);

			PageStackInit();
			break;
		default:
			CSTWARN("Unknown PageID: %08x\n", page_id);
			break;
		}
	}

	CSTCard_Stop();

	CSTStartWindow_ReleaseResource();
	CSTSurveyWindow_ReleaseResource();
	CSTEndWindow_ReleaseResource();

	CSTDirectFBRelease();

	return 0;
}

