#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "cst.h"
#include "cst_storage.h"

static char *fs_types[] = {
	"vfat",
	"msdos",
	"ext3",
	"ext2",
};

static mount_info_t minfo;

static int
CSTStorage_FindDisk(void)
{
	struct stat filestat;
	char path[FILENAME_MAX];
	int ret;
	int i = 0;

	do {
		sprintf(path, "/sys/block/sd%c", 'a' + i);
		ret = stat(path, &filestat);
	} while (ret != 0 && ++i < NR_SUPPORT_DISK);

	if (ret == 0)
		return i;

	return -1;
}

static int
CSTStorage_RemountDisk(mount_info_t *minfo, int flag)
{
	char path[FILENAME_MAX];
	int mount_flags = MS_REMOUNT;
	int ret;

	if (minfo->disk == -1)
		return -1;

	if ((flag & CST_MNT_RDONLY) == CST_MNT_RDONLY)
		mount_flags = MS_REMOUNT | MS_RDONLY;

	sprintf(path, "/dev/sd%c1", 'a' + minfo->disk);

	ret = mount(path, MOUNT_DIR, minfo->fs, mount_flags, NULL);

	return ret;
}

static int
CSTStorage_MountDisk(mount_info_t *minfo, int flag)
{
	char path[FILENAME_MAX];
	int mount_flags;
	int ret;
	int i = 0;

	if (minfo->disk == -1)
		return -1;

	if ((flag & CST_MNT_RDWR) == CST_MNT_RDWR)
		mount_flags = 0;
	else
		mount_flags = MS_RDONLY;

	sprintf(path, "/dev/sd%c1", 'a' + minfo->disk);

	do {
		ret = mount(path, MOUNT_DIR, fs_types[i], mount_flags, NULL);
		if (ret < 0 && errno == EBUSY) {
			ret = mount(path, MOUNT_DIR, fs_types[i],
				    mount_flags | MS_REMOUNT, NULL);
		}
		if (ret == 0) {
			minfo->fs = fs_types[i];
			return 0;
		}
	} while (fs_types[++i]);

	return -1;
}

static void
CSTStorage_UnmountDisk(void)
{
	umount(MOUNT_DIR);
}

static int
CSTStorage_CheckDisk(mount_info_t *minfo)
{
	char path[FILENAME_MAX];
	struct stat filestat;
	int ret;

	if (minfo->disk == -1)
		return -1;

	sprintf(path, "/sys/block/sd%c", 'a' + minfo->disk);
	ret = stat(path, &filestat);

	return ret;
}

int
CSTStorage_CheckFile(char *file)
{
	int ret = 0;
	struct stat filestat;

	if (minfo.disk != -1) {
		ret = CSTStorage_CheckDisk(&minfo);
		if (ret == 0)
			ret = CSTStorage_RemountDisk(&minfo, CST_MNT_RDONLY);
		else {
			CSTStorage_UnmountDisk();
			minfo.disk = -1;
		}
	}

	if (minfo.disk == -1 || ret != 0) {
		minfo.disk = CSTStorage_FindDisk();
		if (minfo.disk != -1)
			ret = CSTStorage_MountDisk(&minfo, CST_MNT_RDONLY);
	}

	if (minfo.disk != -1 && ret == 0) {
		ret = stat(file, &filestat);
		if (ret == 0)
			return 1; /* exist */
		else
			return 0;
	}
	return -1;/* Can't find storage */
}

int
CSTStorage_SaveFile(char *file)
{
	char cmd[FILENAME_MAX];
	int ret = 0;

	if (minfo.disk != -1) {
		ret = CSTStorage_CheckDisk(&minfo);
		if (ret == 0)
			ret = CSTStorage_RemountDisk(&minfo, CST_MNT_RDWR);
		else {
			CSTStorage_UnmountDisk();
			minfo.disk = -1;
		}
	}

	if (minfo.disk == -1 || ret != 0) {
		minfo.disk = CSTStorage_FindDisk();
		if (minfo.disk != -1)
			ret = CSTStorage_MountDisk(&minfo, CST_MNT_RDWR);
	}

	if (minfo.disk != -1 && ret == 0) {
		/* copy to disk from work */
		sprintf(cmd, "mkdir -p %s", STORAGE_DIR);
		system(cmd);

		sprintf(cmd, "cp %s %s", file, STORAGE_DIR);
		system(cmd);
		CSTDEBUG("CMD(%s)\n", cmd);
		CSTINFO("SAVE(%s)\n", file);

		CSTStorage_RemountDisk(&minfo, CST_MNT_RDONLY);
		return 0;
	}

	CSTWARN("Warning: Can't find Storage.\n");
	minfo.disk = -1;
	return -1;
}

int
CSTStorage_Init(void)
{
	int ret;
	int i = 0;

	do {
		minfo.disk = i;
		ret = CSTStorage_MountDisk(&minfo, CST_MNT_RDONLY);
	} while (ret != 0 && ++i < NR_SUPPORT_DISK);

	if (ret < 0)
		minfo.disk = -1;

	return ret;
}

#if defined(DEBUG_STANDALONE)
int
main(int argc, char **argv)
{
	if (argc < 2)
		return 0;

	CSTStorage_Init();

	CSTDEBUG("Currentdisk: %d\n", minfo.disk);

	CSTStorage_SaveFile(argv[1]);

	return 0;
}
#endif
