#include <sys/stat.h> 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cst_survey.h"

#define FILE_BUF_SIZE (65536)

static SSheet_t *SurveyInfoHead;

static char *
CSTSurvey_GetLineHead(char *buf)
{
	char *ptr;

	if (!buf)
		return NULL;

	ptr = buf;
	while (1) {
		ptr += strcspn(ptr, "\r\n");
		ptr += strspn(ptr, " \t\r\n");
		switch (*ptr) {
		case '\0':
			return NULL;
		case '#':
			continue;
		default:
			return ptr;
		}
	}
}

static char *
CSTSurvey_SearchString(char *buf, char *str)
{
	char *ptr;
	size_t len = strlen(str);
	int ret;

	if (!len)
		return NULL;

	ptr = buf;
	while (1) {
		ptr = CSTSurvey_GetLineHead(ptr);
		if (!ptr)
			return NULL;

		ret = strncmp(ptr, str, len);
		if (!ret)
			return (ptr + len);
	}
}

static char *
CSTSurvey_GetQAParam(char *ptr, SQAParam_t *param)
{
	while (1) {
		switch (*(ptr++)) {
		case 'S':
			param->flag |= QAFLAG_SINGLE;
			break;
		case 'M':
			param->flag |= QAFLAG_MULTI;
			break;
		case 'N':
			param->flag |= QAFLAG_NEXT;
			break;
		case 'J':
			{
				int sheet;
				if (sscanf(ptr, "%d", &sheet) == 1) {
					param->sheet = sheet;
					param->flag |= QAFLAG_JUMP;
					ptr += strspn(ptr, "0123456789");
				}
			}
			break;
		case ']':
			return ptr++;
		default:
			/* Not support */
			return NULL;
		}
	}
}

static char *
CSTSurvey_GetQAString(char *ptr, SQA_t *qa)
{
	size_t len;
	ptr += strspn(ptr, ":");
	len = strcspn(ptr, "\r\n");

	if (len > QA_MSG_SIZE - 1)
		len = QA_MSG_SIZE - 1;

	memcpy(qa->msg, ptr, len);
	qa->msg[len] = '\0';

	return ptr;
}

static int
CSTSurvey_ParseQA(char *buf, SSheet_t *sheet)
{
	char *ptr;
	int ans = 0;

	ptr = buf;

	while (1) {
		ptr = CSTSurvey_GetLineHead(ptr);
		if (!ptr)
			return -1;
		if (strncmp(ptr, "Question[", 9) == 0) {
			ptr += 9;
			ptr = CSTSurvey_GetQAParam(ptr,
						   &sheet->question.param);
			if (!ptr)
				return -1;
			ptr = CSTSurvey_GetQAString(ptr, &sheet->question);
			if (!ptr)
				return -1;
		} else if (strncmp(ptr, "Default[", 8) == 0) {
			ptr += 8;
			ptr = CSTSurvey_GetQAParam(ptr,
						   &sheet->default_action);
			if (!ptr)
				return -1;
		} else if (strncmp(ptr, "Answer[", 7) == 0) {
			ptr += 7;
			ptr = CSTSurvey_GetQAParam(ptr,
						   &sheet->answer[ans].param);
			if (!ptr)
				return -1;
			ptr = CSTSurvey_GetQAString(ptr,
						    &sheet->answer[ans]);
			if (!ptr)
				return -1;
			ans++;
		} else if (strncmp(ptr, "}", 1) == 0) {
			sheet->nr_answer = ans;
			ptr += 1;
			return 0;
		}
	}
	return 0;
}

#if defined(DEBUG)
static void
CSTSurvey_DumpSheet(SSheet_t *sheet)
{
	int i;

	printf("ID: %d (%p)\n", sheet->id, sheet);
	printf("\tnext      : %p\n", sheet->next);
	printf("\tQuestion  : <%08x> %s\n",
	       sheet->question.param.flag, sheet->question.msg);
	printf("\tDefault   : <%08x>\n", sheet->default_action.flag);
	for (i=0; i<sheet->nr_answer; i++) {
		printf("\tAnswer[%2d]: <%08x> %s\n",
		       i, sheet->answer[i].param.flag, sheet->answer[i].msg);
	}
}
#endif

static int
CSTSurvey_RegisterSheet(SSheet_t *sheet)
{
	SSheet_t *ptr;
	SSheet_t *cur;

	ptr = (SSheet_t *)malloc(sizeof(SSheet_t));
	if (!ptr)
		return -1;

	memcpy(ptr, sheet, sizeof(SSheet_t));

	if (!SurveyInfoHead)
		SurveyInfoHead = ptr;
	else {
		cur = SurveyInfoHead;
		while (cur->next)
			cur = cur->next;

		cur->next = ptr;
	}

	return 0;
}

void
CSTSurvey_UnregisterSheet(void)
{
	SSheet_t *cur = SurveyInfoHead;
	SSheet_t *next;

	while (cur) {
		next = cur->next;
		free(cur);
		cur = next;
	}
	SurveyInfoHead = NULL;
}

int
CSTSurvey_ParseFile(char *file)
{
	FILE *fp;
	char *buf;
	char *ptr;
	struct stat filestat;
	size_t filesize;
	int ret;

	SSheet_t sheet;

	ret = stat(file, &filestat);
	if (ret < 0)
		return -1;

	filesize = (size_t)filestat.st_size;
	buf = (char *)malloc(filesize);
	if (!buf)
		return -1;

	fp = fopen(file, "rb");
	fread(buf, filesize, 1, fp);
	fclose(fp);

	ptr = buf;
	while (1) {
		int sheet_id;
		char start_char;
		ptr = CSTSurvey_SearchString(ptr, "Sheet");
		if (!ptr)
			break;

		ret = sscanf(ptr, "%d%c", &sheet_id, &start_char);
		if (ret < 2)
			continue;
		if (sheet_id <= 0)
			continue;
		if (start_char == '{') {
			ptr += strspn(ptr, "0123456789{");
			memset(&sheet, 0, sizeof(sheet));
			sheet.id = sheet_id;
			ret = CSTSurvey_ParseQA(ptr, &sheet);
			if (ret)
				continue;
			ret = CSTSurvey_RegisterSheet(&sheet);
		}
	}

	free(buf);
	return 0;
}

SSheet_t *
CSTSurvey_GetSheetInfo(unsigned int sheet_id)
{
	SSheet_t *cur = SurveyInfoHead;
	while (cur) {
		if (cur->id == sheet_id)
			return cur;

		cur = cur->next;
	}
	return NULL;
}

int
CSTSurvey_GetTopSheetID(void)
{
	SSheet_t *top;
  	SSheet_t *cur = top = SurveyInfoHead;
	while (cur) {
		if (cur->id < top->id)
			top = cur;

		cur = cur->next;
	}
	return top->id;
}

#if defined(DEBUG_STANDALONE)
int
main(int argc, char **argv)
{
	SSheet_t *sheet;

	/* Parse & Registration */
	CSTSurveyParseFile("../data/survey.txt");

	/* Test */
	sheet = CSTSurvey_GetTopSheetInfo();
	if (sheet)
		CSTSurvey_DumpSheet(sheet);

	sheet = CSTSurvey_GetSheetInfo(3);
	if (sheet)
		CSTSurvey_DumpSheet(sheet);

	/* Deregistration */
	CSTSurvey_UnregisterSheet();
	return 0;
}
#endif
